"""
This class takes a diff and creates SQL statements for each diff type.

Author: Romke Jonker
Email: romke@rnadesign.net
"""

from .sql_creator import InsertSqlCreator, UpdateSqlCreator, DeleteSqlCreator
from ..compiler.alias_compiler import AliasCompiler


class DiffToSql:

    def __init__(self):
        self._insert_sql_creator = InsertSqlCreator()
        self._update_sql_creator = UpdateSqlCreator()
        self._delete_sql_creator = DeleteSqlCreator()

    def diff_sql(self, mapping, diffs, context=None):
        # get from tuple
        inserts, updates, deletes = diffs

        # add alias and parameter names to mapping before creating sql
        aliased_mapping = AliasCompiler().compile(mapping)

        # create sql for each diff
        insert_sql = list(self._insert_sql_creator.create_sql(aliased_mapping, inserts, context))
        update_sql = list(self._update_sql_creator.create_sql(aliased_mapping, updates, context))
        delete_sql = list(self._delete_sql_creator.create_sql(aliased_mapping, deletes, context))

        return insert_sql + update_sql + delete_sql



"""
This class compiles a mapping into a list of parameters for each column.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
from itertools import chain


class ParametersCompiler:

    def compile(self, mapping):
        if 'columns' not in mapping:
            return ''
        # list of tuples. Use tuples because they are hashable, so we can use them as keys in a dictionary
        columns = [tuple(self._column(c)) for c in mapping['columns']]
        return columns

    def _column(self, column):
        return self._attributes(column['attributes'], True)

    def _attributes(self, attributes, is_root_table=False):
        return chain(*[self._attribute(a, is_root_table) for a in attributes])

    def _attribute(self, attribute, is_root_table):
        if 'foreign-key' in attribute:
            foreign_key = attribute['foreign-key']
            return self._attributes(foreign_key['attributes'])

        parameter = attribute['parameter']
        return [parameter]



import logging
import re
from abc import ABC, abstractmethod
from enum import Enum

import pandas as pd

'''
This class allows for different execution styles. 
In particular, it allows for dependent queries, where the result of the first query is used as a parameter in the second query.
'''

_logger = logging.getLogger(__name__)


class QueryExecutor(ABC):
    def __init__(self, line_number, operation_type, table_name, context):
        self.line_number = line_number
        self.operation_type = operation_type
        self.table_name = table_name
        self.context = context

    @abstractmethod
    def execute(self, cursor):
        pass

    '''
    Called when user sets execute=False. Purpose is to inspect the queries and parameters that would be executed.
    '''
    @abstractmethod
    def fake_execute(self):
        pass

    @abstractmethod
    def queries(self):
        pass

    def _replace_placeholders(self, query):
        # replace :xyz with %(xyz)s using regex
        return re.sub(r':(\w+)', r'%(\1)s', query)


class SimpleQueryExecutor(QueryExecutor):
    def __init__(self, line_number, operation_type, table_name, query, params, context):
        super().__init__(line_number, operation_type, table_name, context)
        self.query = query
        self.params = params

    def queries(self):
        return [(self.query, self.params)]

    def execute(self, cursor):
        # replace ':' style place holders with '%' style
        psycopg_query = self._replace_placeholders(self.query)
        # replace NA values with None in params dictionary
        params_with_none = {k: None if pd.isna(v) else v for k, v in self.params.items()}

        try:
            # execute query
            cursor.execute(psycopg_query, params_with_none)
        except Exception as e:
            error = str(e)
            return ExecutionResult(self.line_number, self.operation_type, False, 0, self.table_name, self.query, self.params, self.context, error=error)

        # Get the number of affected rows
        rowcount = cursor.rowcount

        # verify row was affected
        if rowcount == 0:
            error = 'No row was affected'
            return ExecutionResult(self.line_number, self.operation_type, False, rowcount, self.table_name, self.query, self.params, self.context, error=error)

        # verify no more than one row was affected
        if rowcount > 1:
            # we must not commit the transaction
            error = "More than one row was affected, do not commit."
            return ExecutionResult(self.line_number, self.operation_type, False, rowcount, self.table_name, self.query, self.params, self.context, error, True)

        result = ExecutionResult(self.line_number, self.operation_type, True, rowcount, self.table_name, self.query, self.params, self.context)
        return result

    def fake_execute(self):
        return ExecutionResult(self.line_number, self.operation_type, False, 0, self.table_name, self.query, self.params, self.context)

# This class is used to store input rows that failed to compile into a query
class FailedQueryExecutor(QueryExecutor):
    def __init__(self, line_number, operation_type, table_name, context, error):
        super().__init__(line_number, operation_type, table_name, context)
        self.error = error

    def queries(self):
        return []

    def execute(self, cursor):
        return ExecutionResult(self.line_number, self.operation_type, False, 0, self.table_name, None, {}, self.context, error=self.error)

    def fake_execute(self):
        return ExecutionResult(self.line_number, self.operation_type, False, 0, self.table_name, None, {}, self.context, error=self.error)

'''
This class allows for dependent queries, where the result of the first query is used as a parameter in the second query.
This is useful for extensions, such as the ir_model_data table in Odoo.
'''
class DependentQueryExecutor(QueryExecutor):
    def __init__(self, line_number, operation_type, table_name, context, initial_query, dependent_query):
        super().__init__(line_number, operation_type, table_name, context)
        self.query = initial_query[0]
        self.params = initial_query[1]
        self.dependent_query = dependent_query


    def queries(self):
        return [(self.query, self.params), self.dependent_query]

    def execute(self, cursor):
        # replace ':' style place holders with '%' style
        query_0 = self._replace_placeholders(self.query)

        # replace NA values with None in params dictionary
        params_0 = {k: None if pd.isna(v) else v for k, v in self.params.items()}

        cursor.execute(query_0, params_0)
        result = cursor.fetchone()
        rowcount_0 = cursor.rowcount

        # verify row was affected
        if rowcount_0 == 0:
            _logger.warning("No row was affected. Query: %s, Params: %s" % (query_0, params_0))
            execution_result = ExecutionResult(self.line_number, self.operation_type, False, rowcount_0, self.table_name, self.query, params_0, self.context)
            # skip execution of dependent query
            return execution_result

        # verify no more than one row was affected
        if rowcount_0 > 1:
            # raise exception, bec/ we must not commit the transaction
            raise ValueError("More than one row was affected. Inserts: %s, Query: %s, Params: %s" % (rowcount_0, query_0, params_0))

        execution_result = ExecutionResult(self.line_number, self.operation_type, True, rowcount_0, self.table_name, self.query, params_0, self.context)

        # replace ':' style place holders with '%' style
        query_1 = self.dependent_query[0]
        params_1 = self.dependent_query[1]
        params_1['res_id'] = result[0]

        cursor.execute(self._replace_placeholders(query_1), params_1)
        rowcount_1 = cursor.rowcount

        # create execution result of dependent query
        dependent_execution_result = ExecutionResult(self.line_number, self.operation_type, True, rowcount_1, self.table_name, query_1, params_1, self.context)
        execution_result.dependent_execution_result = dependent_execution_result

        return execution_result

    def fake_execute(self):
        execution_result = ExecutionResult(self.line_number, self.operation_type, False, 0, self.table_name, self.query, self.params, self.context)
        dependent_execution_result = ExecutionResult(self.line_number, self.operation_type, True, 0, self.table_name, self.dependent_query[0], self.dependent_query[1], self.context)
        execution_result.dependent_execution_result = dependent_execution_result
        return execution_result



class OperationType(Enum):
    INSERT = "INSERT"
    UPDATE = "UPDATE"
    DELETE = "DELETE"

    def __str__(self):
        return self.value

    def __repr__(self):
        return "'" + self.value + "'"


class ExecutionResult:
    def __init__(self, line_number, operation_type, success, rowcount, table_name, query, params, context, error=None, block_commit=False):
        # convert numpy int64 to int
        self.line_number = int(line_number) if line_number is not None else None
        if not isinstance(operation_type, OperationType):
            raise ValueError(f"operation_type must be an instance of OperationType Enum, not {type(operation_type)}")
        self.operation_type = operation_type
        self.success = success
        self.rowcount = rowcount
        self.table_name = table_name
        self.query = query
        self.params = params
        self.context = context
        self.error = error
        self.block_commit = block_commit
        self.dependent_execution_result = None

    def __str__(self):
        return f"Type: {self.operation_type}, Success: {self.success}, Rowcount: {self.rowcount}, Table: {self.table_name}, Query: {self.query}, Params: {self.params}"

    def report(instance, execute):
        if execute:
            return [{key: value for key, value in vars(instance).items() if value is not None and key not in ['block_commit', 'dependent_execution_result']}]
        else:
            return [{key: value for key, value in vars(instance).items() if value is not None and key not in ['block_commit', 'success', 'rowcount', 'dependent_execution_result']}]



"""
This class provides the main database service to list tables, create mappings, retrieve table contents and to post updates.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
import importlib.util
import logging
import os
import re
import sys
from io import StringIO
from itertools import chain

import pandas as pd
import psycopg2
from pandas import DataFrame

from stimula.compiler.header_compiler import HeaderCompiler
from stimula.compiler.select_compiler import SelectCompiler
from stimula.compiler.types_compiler import TypesCompiler
from stimula.header.csv_header_parser import HeaderParser
from stimula.header.header_lexer import HeaderLexer
from stimula.header.header_merger import HeaderMerger
from stimula.header.odoo_header_parser import OdooHeaderParser
from .context import cnx_context, get_metadata
from .diff_to_sql import DiffToSql
from .query_executor import OperationType
from .reporter import Reporter
from ..compiler.alias_compiler import AliasCompiler

_logger = logging.getLogger(__name__)


class DB:
    def __init__(self):
        self._lexer = HeaderLexer()
        self._diff_to_sql = DiffToSql()

    def get_tables(self, filter=None):

        cr = cnx_context.cr

        # query to get all tables, excluding views
        query = "SELECT table_name FROM information_schema.tables WHERE table_schema = 'public' AND table_type = 'BASE TABLE'"

        cr.execute(query)

        # fetch all rows
        all_tables = cr.fetchall()

        if filter is None:
            # return all tables
            unsorted = [t for t in all_tables]
        else:
            # use filter as a regular expression
            regex = re.compile(filter)
            unsorted = [t for t in all_tables if regex.search(t[0])]

        # get row counts
        with_counts = self.get_counts(unsorted)

        # return sorted(with_count)
        sorted_tables = sorted(with_counts, key=lambda e: e[0])

        # convert to dictionary for JSON serialization
        return [{'name': table_name, 'count': row_count} for table_name, row_count in sorted_tables]

    def get_counts(self, tables):
        cr = cnx_context.cr
        for table in tables:
            query = f"SELECT COUNT(*) FROM {table[0]}"
            cr.execute(query)
            row_count = cr.fetchone()[0]
            yield table[0], row_count

    def get_count(self, table_name, header, where_clause):
        # create metadata object from context
        metadata = get_metadata(cnx_context.cnx)

        if not header:
            # if header parameter is empty, then use the default header
            default_mapping = OdooHeaderParser(metadata, cnx_context.cr).parse(table_name)

            # remove columns that are not enabled
            mapping = {**default_mapping, 'columns': [c for c in default_mapping['columns'] if c.get('enabled', False)]}

        else:
            # parse header to build syntax tree
            mapping = HeaderParser(metadata, table_name).parse_csv(header)

        query = SelectCompiler().compile_count_query(mapping, where_clause)

        cr = cnx_context.cr

        # execute query
        cr.execute(query)

        # fetch result
        result = cr.fetchone()

        return result[0]

    def get_table(self, table_name, header, where_clause=None):
        # create metadata object from context
        metadata = get_metadata(cnx_context.cnx)

        if not header:
            # if header parameter is empty, then use the default header
            default_mapping = OdooHeaderParser(metadata, cnx_context.cr).parse(table_name)

            # remove columns that are not enabled
            mapping = {**default_mapping, 'columns': [c for c in default_mapping['columns'] if c.get('enabled', False)]}

        else:
            # parse header to build syntax tree
            mapping = HeaderParser(metadata, table_name).parse_csv(header)

        df = self._read_from_db(mapping, where_clause)

        return df, mapping

    def get_table_as_csv(self, table_name, header, where_clause, escapechar=''):
        # get table as dataframe
        df, mapping = self.get_table(table_name, header, where_clause)

        return self.convert_to_csv(df, mapping, escapechar)

    def convert_to_csv(self, df, mapping, escapechar=''):
        # need converters from types compiler
        column_names = HeaderCompiler().compile_list(mapping)
        column_types = TypesCompiler().compile(mapping, column_names)

        # get converters from column_types
        converters = column_types['write_csv_converters']

        # apply converters to dataframe
        for column in column_names:
            if column in converters:
                df[column] = df[column].apply(converters[column])

        # convert dataframe to csv
        return df.to_csv(index=False, escapechar=escapechar)

    def post_table_get_diff(self, table_name, header, where_clause, body, skiprows=0, insert=False, update=False, delete=False, execute=False, commit=False, post_script=None, context=None):
        # create diffs and sql
        diffs, sql = self._get_diffs_and_sql(table_name, header, where_clause, body, skiprows, insert, update, delete, post_script, context)
        # if execute
        if execute:
            # execute sql statements
            results_tuple = self._execute_sql(sql, True, commit)
            # concatenate results
            results = [result for results in results_tuple for result in results]
            # zip row counts with diffs statements
            for diff, result in zip(diffs, results):
                diff.append(result.rowcount)

        return diffs

    def post_table_get_sql(self, table_name, header, where_clause, body, skiprows=0, insert=False, update=False, delete=False, execute=False, commit=False, post_script=None, context=None):
        # create diffs and sql
        _, query_executors = self._get_diffs_and_sql(table_name, header, where_clause, body, skiprows, insert, update, delete, post_script, context)

        # execute sql statements, or fake if execute is false
        sqls = self._execute_sql(query_executors, execute, commit)

        # convert sql to dataframe
        return self._convert_to_df(sqls, execute)

    def post_table_get_full_report(self, table_name, header, where_clause, body, skiprows=0, insert=False, update=False, delete=False, execute=False, commit=False,
                                   post_script=None, context=None):
        # create diffs and sql
        diff, query_executors = self._get_diffs_and_sql(table_name, header, where_clause, body, skiprows, insert, update, delete, post_script, context)

        # execute sql statements
        execution_results = self._execute_sql(query_executors, execute, commit)

        # create full report
        return Reporter().create_post_report([table_name], [body], [context], execution_results, execute, commit, skiprows)

    def post_multiple_tables_get_full_report(self, table_names, header, where_clause, contents, skiprows=0, insert=False, update=False, delete=False, execute=False, commit=False,
                                             post_script=None, context=None):
        assert len(table_names) == len(contents), f"Provide exactly one file for each table name, so {len(table_names)}, not {len(contents)}"
        assert header is None, "Header must be None when posting multiple tables"
        assert skiprows ==1, "Skiprows must be 1 when posting multiple tables"
        assert post_script is None, "Post script must be None when posting multiple tables"
        assert context is not None and len(context) == len(table_names), "Provide exactly one context for each table name, not %s" % len(context or [])

        query_executors = []

        # Iterate over tables here.
        for table_name, file_context, content in zip(table_names, context, contents):
            # decode binary content
            text_content = content.decode('utf-8')

            # get header from first line
            header = text_content.split('\n', 1)[0]

            # create diffs and sql
            _, qe = self._get_diffs_and_sql(table_name, header, where_clause, text_content, skiprows, insert, update, delete, post_script, file_context)
            query_executors.extend(qe)

        # execute sql statements
        execution_results = self._execute_sql(query_executors, execute, commit)

        # create full report
        return Reporter().create_post_report(table_names, contents, context, execution_results, execute, commit, skiprows)


    def _convert_to_df(self, sqls, showResult):
        # create empty pandas dataframe. First column contains sql
        if showResult:
            result = pd.DataFrame(columns=['rows', 'sql'])
        else:
            result = pd.DataFrame(columns=['sql'])

        # iterate sql statements
        index = 0
        for er in sqls:
            # create dictionary with query and value_dict
            value_dict = {**er.params, 'sql': er.query}
            if showResult:
                value_dict = {**value_dict, 'rows': er.rowcount}
            # append a new row to the bottom result with values from value dictionary using concat
            result = pd.concat([result, pd.DataFrame(value_dict, index=[index])], ignore_index=True)
            index += 1

            # append dependent query result
            dep_er = er.dependent_execution_result
            if dep_er:
                value_dict = {**dep_er.params, 'sql': dep_er.query}
                if showResult:
                    value_dict = {**value_dict, 'rows': dep_er.rowcount}
                result = pd.concat([result, pd.DataFrame(value_dict, index=[index])], ignore_index=True)
                index += 1

        # force pandas to not convert int columns to float if they contain NaNs
        result = result.convert_dtypes()

        # return dataframe
        return result

    def post_table_get_summary(self, table_name, header, where_clause, body, skiprows=0, insert=False, update=False, delete=False, execute=False, commit=False):
        with self._database_session() as (engine, metadata):
            # parse header to build syntax tree
            diffs, sql = self._get_diffs_and_sql(engine, metadata, table_name, header, where_clause, body, skiprows, insert, update, delete)

            # create result object
            result = {}

            if execute:
                # execute sql statements
                results_tuple = self._execute_sql(engine, sql, commit)
                # summarize results
                # zip row counts with sql statements

            return result

    def _get_diffs_and_sql(self, table_name, header, where_clause, body, skiprows, insert, update, delete, post_script, context):
        # get cnx from context
        cnx = cnx_context.cnx

        # if header is empty and skiprows is 1, then take the first line as header
        if not header and skiprows == 1:
            # get header from first line
            header = body.split('\n', 1)[0]

        # assert that we have a header
        assert header, "Header is required, either as a parameter or as the first line in the body with skiprows set to 1"

        # parse header to build syntax tree
        mapping = HeaderParser(get_metadata(cnx), table_name).parse_csv(header)

        # read dataframe from request first, so we can give feedback on errors in the request
        df_request = self._read_from_request(mapping, body, skiprows, post_script)

        # read dataframe from DB
        df_db = self._read_from_db(mapping, where_clause, set_index=True)

        # compare to obtain diff
        diffs = self._compare(df_request, df_db, insert, update, delete)

        # create sql statements and parameters
        sqls = self._diff_to_sql.diff_sql(mapping, diffs, context)

        return diffs, sqls

    def _read_from_request(self, mapping, body, skiprows, post_script=None):
        # get columns and unique columns.
        column_names = HeaderCompiler().compile_list(mapping, include_skip=True)
        index_columns = HeaderCompiler().compile_list_unique(mapping)
        column_types = TypesCompiler().compile(mapping, column_names, include_skip=True)
        deduplicate_columns = HeaderCompiler().compile_list_deduplicate(mapping)

        # assert that at least one column header is not empty
        if not [c for c in column_names if c != '']:
            raise ValueError("At least one column header must not be empty")

        # read body as csv to count the number of columns in the body
        # there may be a more efficient way without having to read the csv twice.
        body_column_count = self._count_body_columns(body)

        # pad column names with empty columns if the body has more columns than the header
        if body_column_count > len(column_names):
            # add empty columns to the header
            column_names += [''] * (body_column_count - len(column_names))

        # replace empty column names with skip, skip1, skip2. This is because pandas requires column names to be unique
        non_empty_column_names = list(self._replace_empty_columns_with_skip(column_names))

        # find duplicate column names
        duplicate_column_names = self._find_duplicate_names(non_empty_column_names)

        # if there are duplicate column names, raise an exception
        if duplicate_column_names:
            raise ValueError(f"Duplicate column names are not supported: {', '.join(duplicate_column_names)}")

        # get list of columns to use in the output dataframe
        use_columns = [c for c in non_empty_column_names if c in column_names]

        # list names of columns with datetime64 or date type, because we need to parse them as datetime
        parse_dates = column_types.get('read_csv_parse_dates', {})

        # get converter dictionary
        converters = column_types.get('read_csv_converters', {})

        # get dtypes for read_csv
        dtype = column_types.get('read_csv_dtypes', {})

        # create initial column names to read the csv before padding
        initial_index_columns = [c for c in index_columns if c in non_empty_column_names[:body_column_count]]
        initial_names = non_empty_column_names[:body_column_count]
        initial_usecols = [c for c in use_columns if c in non_empty_column_names[:body_column_count]]

        # read csv from request body
        # treat '' as missing value, but treat NA as string
        df = pd.read_csv(
            StringIO(body),
            names=initial_names,
            index_col=initial_index_columns,
            skipinitialspace=True,
            skiprows=skiprows,
            # skip the empty column names that we renamed to skip, skip1, etc
            usecols=initial_usecols,
            parse_dates=parse_dates,
            converters=converters,
            na_values=[''],
            keep_default_na=False
        )

        # iterate columns types and convert to the correct type.
        # Need to do this after reading, because pd sets converter results to type object
        for column, type in dtype.items():
            # skip setting index types
            if column in df.columns:
                # convert column to the correct type
                df[column] = df[column].astype(type)

        df_padded = self._pad_dataframe_with_empty_columns(df, use_columns, index_columns, converters)

        # evaluate column expressions
        self._evaluate_expressions(df_padded, mapping, index_columns)

        # insert a column with line numbers
        df_padded.insert(0, '__line__', range(0, len(df)))

        # deduplicate if requested
        if deduplicate_columns:
            df_padded = self._deduplicate(df_padded, deduplicate_columns)

        # verify that there are no duplicate index values
        if df_padded.index.has_duplicates:
            # find duplicate index values
            duplicates = df_padded.index[df_padded.index.duplicated()]
            # convert to name-value dict
            duplicate_map = {k: v for k, v in zip(duplicates.names, duplicates.values)}

            raise ValueError(f"Duplicates found: {duplicate_map}")


        # apply post script if provided
        if post_script:
            # execute post script
            self._execute_post_script(df_padded, post_script)

        return df_padded

    def _evaluate_expressions(self, df, mapping, index_columns):
        # a column header can contain a python expression. Evaluate these now that we've read all values from CSV.

        # drop index, so that we can use index columns in expressions
        if index_columns:
            df.reset_index(inplace=True)

        # store original column names so we can restore them later
        original_column_names = df.columns

        # remove foreign keys and modifiers in column names so we can use the bare names in expressions
        df.columns = df.columns.str.replace(r'\[.*\]', '', regex=True)
        df.columns = df.columns.str.replace(r'\(.*\)', '', regex=True)

        # now that we have all data, we can evaluate column expressions.
        for column in mapping['columns']:
            if 'exp' in column:
                # get column name, assuming a single attribute
                column_name = column['attributes'][0]['name']

                # evaluate expression of the form <target_column>=<expression>
                expression = f"{column_name}={column['exp']}"
                df.eval(expression, inplace=True)

        # restore column names
        df.columns = original_column_names

        # restore index
        if index_columns:
            df.set_index(index_columns, inplace=True)

        # remove skip columns, because we've evaluated expressions so we no longer need them. Match column name with 'skip=true':
        df.drop(columns=[column for column in df.columns if 'skip=true' in column], errors='ignore', inplace=True)

    def _read_from_db(self, mapping, where_clause, set_index=False):
        # get sqlalchemy engine from context
        engine = cnx_context.engine

        # get enabled and unique columns and column types
        column_names = HeaderCompiler().compile_list(mapping)
        index_columns = HeaderCompiler().compile_list_unique(mapping)
        column_types = TypesCompiler().compile(mapping, column_names)

        # read dataframe from DB
        query = self._create_select_query(mapping, where_clause)
        df = pd.read_sql_query(query, engine)

        # set headers, they must equal the request headers for comparison
        df.columns = column_names

        # set index columns, if header contains unique columns
        if set_index and index_columns:
            df.set_index(index_columns, inplace=True)

        # apply converters after reading from DB, because read_sql_query() doesn't support converters
        converters = column_types['read_db_converters']
        self._apply_converters(column_names, converters, df)

        # force pandas to not convert int columns to float if they contain NaNs
        df = df.convert_dtypes()

        return df

    def _apply_converters(self, column_names, converters, df):
        # iterate the columns to apply converters one by one to either value or index column
        for column in column_names:
            if column in converters:
                if column in df.columns:

                    # apply this converter to value column
                    df[column] = df[column].apply(converters[column])
                elif column in df.index.names:

                    # get the index column
                    index_column = df.index.levels[df.index.names.index(column)]

                    # apply this converter to the index column
                    df.index = df.index.set_levels(index_column.map(converters[column]), level=column)

    def _compare(self, df_request, df_db, insert, update, delete):
        # remove columns with empty names. Don't do this when reading from DB, because in get_table request we also want empty columns
        df_db = df_db.drop(columns=[''], errors='ignore')

        # find rows to insert
        inserted_indices = set(df_request.index.values).difference(df_db.index.values)
        inserts = df_request[df_request.index.isin(inserted_indices)]
        inserts.reset_index(inplace=True)

        # move __line__ to become the left most column after index reset
        inserts = self._move_line_to_front(inserts)

        # find rows to delete
        deleted_indices = set(df_db.index.values).difference(df_request.index.values)
        deletes = df_db[df_db.index.isin(deleted_indices)]
        deletes.reset_index(inplace=True)

        # find rows to update. Left is from request, right is from database
        left = df_request[~df_request.index.isin(inserted_indices)]
        right = df_db[~df_db.index.isin(deleted_indices)]

        # Sort the right DataFrame to match the index of the left DataFrame
        right_df_sorted = right.reindex(left.index)

        # left (from request) has __line__ column. Create a copy without line numbers for comparison
        left_no_line = left.drop(columns=['__line__'])

        # get mask of cells that are N/A or '' in left and right
        mask = (left_no_line.isna() | left_no_line.eq('')) & (right_df_sorted.isna() | right_df_sorted.eq(''))

        # replace N/A or '' with None using mask to ignore differences in N/A or ''
        left_no_line = left_no_line.mask(mask, None)
        right_df_sorted = right_df_sorted.mask(mask, None)

        # compare remaining rows
        updates = left_no_line.compare(right_df_sorted)

        # insert __line__ column, recovering line numbers from left by index
        updates['__line__'] = left.loc[updates.index]['__line__']

        # reset index for updates to match the insert and deletes
        updates.reset_index(inplace=True)

        # move __line__ to become the left most column
        updates = self._move_line_to_front(updates)

        # return result
        return inserts if insert else DataFrame(), updates if update else DataFrame(), deletes if delete else DataFrame()

    def _move_line_to_front(self, df):
        # move __line__ to become the left most column. This has no real purpose, but it makes the dataframes more readable
        # this must also work with the multi-index dataframes coming from the compare function
        # pop __line__ column, don't use drop because that triggers a warning
        line = df.pop('__line__')
        # insert __line__ column as first column
        df.insert(0, '__line__', line)
        return df


    def get_select_statement(self, table_name, header, where_clause=None):
        # get cnx from context
        cnx = cnx_context.cnx

        # parse header to build mapping
        mapping = HeaderParser(get_metadata(cnx), table_name).parse_csv(header)

        # Read data from the database table using pandas
        query = self._create_select_query(mapping, where_clause)

        return str(query)

    def _create_select_query(self, mapping, where_clause):
        # add aliases and parameter names
        aliased_mapping = AliasCompiler().compile(mapping)

        # translate syntax tree to select query
        return SelectCompiler().compile(aliased_mapping, where_clause)

    def get_header_json(self, table_name, header=None):
        # get mapping, merged if needed
        mapping = self._get_header_mapping(table_name, header)

        # compile into header
        json_header = HeaderCompiler().compile_json(mapping)

        return json_header

    def get_header_csv(self, table_name, header=None):
        # get mapping, merged if needed
        mapping = self._get_header_mapping(table_name, header)

        # compile into header
        csv_header = HeaderCompiler().compile_csv(mapping)

        return csv_header

    def _get_header_mapping(self, table_name, header=None):
        # get a MetaData instance from context
        metadata = get_metadata(cnx_context.cnx)

        # create default mapping from table
        mapping = OdooHeaderParser(metadata, cnx_context.cr).parse(table_name)

        if not header:
            # return full mapping
            return mapping

        # parse header to build syntax tree
        parsed_header = HeaderParser(metadata, table_name).parse_csv(header)

        # merge headers
        merged_mapping = HeaderMerger().merge(mapping, parsed_header)

        return merged_mapping

    def _execute_sql(self, query_executors, execute, commit):
        if not execute:
            # fake execution, return result
            return [qe.fake_execute() for qe in query_executors]

        # get cursor from context
        cr = cnx_context.cr

        # execute queries, rerun until exhausted
        result = self._eat_sleep_repeat(query_executors, cr)

        # commit if requested
        if commit:
            cnx_context.cnx.commit()

            # registry may not be available during unit tests
            if hasattr(cnx_context, 'registry'):
                # registry may not have clear_cache() method
                if hasattr(cnx_context.registry, 'clear_cache'):
                    # invalidate caches to avoid stale values coming from cache
                    cnx_context.registry.clear_cache()
                else:
                    _logger.warning("Registry has no clear_cache() method")

        return result

    def _eat_sleep_repeat(self, query_executors, cr):
        # execute in rounds until no new successful queries are found

        # create result lists
        completed = []
        failed = []

        # copy query executors list
        remaining = query_executors.copy()

        done = False

        while not done:
            new_completed_executors = []
            new_completed_results = []
            # iterate query executors
            for query_executor in remaining:
                # create or replace savepoint
                self.create_savepoint()
                # delegate execution to query executor
                execution_result = query_executor.execute(cr)
                # if successful
                if execution_result.success:
                    # append result to list
                    new_completed_results.append(execution_result)
                    # remove from remaining
                    new_completed_executors.append(query_executor)
                else:
                    # rollback to savepoint
                    self.rollback_to_savepoint()
                    # append to failed list
                    failed.append(execution_result)
            if new_completed_results:
                # append new completed to completed list
                completed.extend(new_completed_results)
                # remove completed executors from remaining
                remaining = [qe for qe in remaining if qe not in new_completed_executors]
                # reset failed list and start again
                failed = []
            else:
                # nothing new completed, we're done
                done = True
        # combine completed and failed lists
        all_results = completed + failed

        # set delete queries apart, because they don't have line numbers
        deleted = [result for result in all_results if result.operation_type == OperationType.DELETE]
        insert_and_updates = [result for result in all_results if result.operation_type != OperationType.DELETE]

        # sort by line_number
        insert_and_updates.sort(key=lambda x: x.line_number)

        # append deleted to the end
        return insert_and_updates + deleted


    def set_context(self, url, password):
        # create psycopg2 connection
        cnx = psycopg2.connect(url, password=password)
        # set connection and cursor in context
        cnx_context.cnx = cnx
        cnx_context.cr = cnx.cursor()

    def create_savepoint(self):
        # create savepoint
        cnx_context.cr.execute("SAVEPOINT stimula_savepoint")

    def rollback_to_savepoint(self):
        # rollback to savepoint
        cnx_context.cr.execute("ROLLBACK TO SAVEPOINT stimula_savepoint")

    def _replace_empty_columns_with_skip(self, column_names):
        # replace empty column names with skip, skip1, skip2. This is because pandas doesn't like empty column names
        i = 0
        for name in column_names:
            if name == '':
                name = 'skip' + ('' if i == 0 else str(i))
                i += 1
            yield name

    def _find_duplicate_names(self, names):
        # find duplicate names
        seen = set()
        duplicates = set()
        for name in names:
            if name in seen:
                duplicates.add(name)
            seen.add(name)
        return duplicates

    def _get_selected_columns(self, table_name, all_columns, header):
        # get cnx from context
        cnx = cnx_context.cnx

        # parse header to build syntax tree
        mapping = HeaderParser(get_metadata(cnx), table_name).parse_csv(header)

        # reconstruct original header from tree
        column_keys = HeaderCompiler().compile_list(mapping)

        # create result list for selected and unselected columns
        result_columns = []

        # initiate list for unselected columns with all default columns
        unselected_columns = all_columns.copy()

        # iterate header columns to retain order of selected items
        for key in column_keys:

            # find column with matching key
            column = next((c for c in unselected_columns if c['key'] == key), None)

            # raise error if if column is not found,
            if column is None:
                raise ValueError(f'Column {key} not found in table {table_name}')

            # set selected is true
            column['selected'] = True

            # remove from unselected columns and add to selected columns
            unselected_columns.remove(column)
            result_columns.append(column)

        # add remaining unselected columns to selected columns
        result_columns.extend(unselected_columns)

        return result_columns

    def _count_body_columns(self, body):
        # skipinitialspace must be true, otherwise it may split on comma's in strings
        df_initial = pd.read_csv(StringIO(body), skipinitialspace=True, header=None)

        # return the number of columns in the body
        return len(df_initial.columns)

    def _pad_dataframe_with_empty_columns(self, df, column_names, index_columns, converters):
        # get current column names
        current_columns = df.columns.tolist()

        # get number of columns to pad, take columns and indices into account
        pad_count = len(column_names) - df.shape[1] - df.index.nlevels

        # if there are too few columns
        if pad_count <= 0:
            # nothing to pad
            return df

        # get columns to pad
        pad_columns = column_names[-pad_count:]

        # pad the DataFrame with empty columns
        _logger.info(f"Padding DataFrame with {pad_count} empty columns")
        current_columns += pad_columns

        # be careful not to set index columns, the columns attribute must only be set to non-index column names
        df_padded = df.reindex(columns=current_columns, fill_value=None)

        # apply converters to padded columns
        for column in pad_columns:
            if column in converters:
                df_padded[column] = df_padded[column].apply(converters[column])

        # set additional index columns
        for column in pad_columns:
            if column in index_columns:
                df_padded.set_index(column, append=True, inplace=True)

        return df_padded

    def _deduplicate(self, df, index_columns):
        # reset index
        df_reset = df.reset_index()

        # Remove duplicate rows based on the index column
        df_unique = df_reset.drop_duplicates(subset=index_columns)

        # Set the column back as the index
        df_final = df_unique.set_index(index_columns)

        return df_final

    def _execute_post_script(self, df, post_script):
        # skip if no post script provided
        if post_script is None:
            return df

        # assert that post_script file exists
        assert os.path.exists(post_script), f"Post script file {post_script} not found"

        # import the post script module
        module_name = 'post_script'
        spec = importlib.util.spec_from_file_location(module_name, post_script)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        # verify that the module has an execute function
        assert hasattr(module, 'execute'), f"Post script module {post_script} must have an execute function"

        # execute the post script
        return module.execute(df)



"""
This class generates insert statements based on a mapping.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
import re
from itertools import chain

from stimula.compiler.select_compiler import SelectCompiler


class InsertCompiler:
    """
        header: 'c1[unique=true], c2(c1), c3(b1), c4(b2(a1))'

        query:
        insert into c (c1, c2, c3, c4)
        select :1, c.c0, b.b0, b1.b0
        from c, b, b as b0
        join a on b0.b2=a.a0
        where c.c1 = :2 and b.b1 = :3 and a.a1=:4
        returning id;
    """

    def compile(self, mapping):
        insert_clause = InsertClauseCompiler().compile(mapping)
        select_clause = SelectClauseCompiler().compile(mapping)
        from_clause = FromClauseCompiler(True).compile(mapping)
        where_clause = ForeignWhereClauseCompiler(True).compile(mapping)
        returning_clause = ReturningClauseCompiler().compile(mapping)

        # this where clause is also used in update query. Here we need 'where'
        if where_clause:
            where_clause = ' where ' + where_clause

        return f'{insert_clause}{select_clause}{from_clause}{where_clause}{returning_clause}'


class InsertClauseCompiler:
    def compile(self, mapping):
        # table name is first argument
        table = mapping['table']

        # get lists of lists of attributes
        column_lists = [self._column(c) for c in mapping['columns']]

        # flatten list of lists
        attribute_list = [a for attributes in column_lists for a in attributes]

        # assert there's at least one attribute to insert
        assert attribute_list, f'No attributes to insert in table {table}'

        # comma separate
        columns = ', '.join(attribute_list)

        return f'insert into {table}({columns})'

    def _column(self, column):
        attributes = self._attributes(column['attributes'])
        return attributes

    def _attributes(self, attributes):
        # skip extensions on base table, because they are not columns. We'll insert them in a separate query
        return [a['name'] for a in attributes if not a.get('foreign-key', {}).get('extension')]


class SelectClauseCompiler:
    def compile(self, mapping):
        # get list of attributes per column
        columns = [self._column(c) for c in mapping['columns']]
        # flatten list of lists
        attributes = [a for attributes in columns for a in attributes]
        # comma separate cells
        return ' select ' + ', '.join(attributes)

    def _column(self, column):
        attributes = self._attributes(column['attributes'])
        return attributes

    def _attributes(self, attributes):
        # iterate attributes to get list of lists.
        # skip extensions on base table, because they are not columns. We'll insert them in a separate query
        return [self._attribute(a) for a in attributes if not a.get('foreign-key', {}).get('extension')]

    def _attribute(self, attribute):
        if 'foreign-key' in attribute:
            foreign_key = attribute['foreign-key']
            # but table names may need an alias
            target_name = foreign_key['table']
            target_alias = foreign_key.get('alias', target_name)
            target_column = foreign_key['name']
            # no need to recurse
            return f'{target_alias}.{target_column}'

        # get value from parameter
        return f':{attribute["parameter"]}'


'''
There's a subtle difference in the from clause between insert and update statements with respect to extensions:
insert query: we must not join with extension table, because we're inserting a new record and we don't have a primary key yet.
update query: we must join with extension table, because we're updating an existing record and we need to filter by the extension table.
'''
class FromClauseCompiler:

    def __init__(self, is_insert_query):
        # if this is an insert query, we must not join with extension tables
        self._is_insert_query = is_insert_query

    def compile(self, mapping):
        # get table
        table = mapping['table']

        # compile columns
        columns = [self._column(c, table) for c in mapping['columns']]

        # filter out empty columns that don't need a from clause
        filtered_columns = [column for column in columns if column]
        if not filtered_columns:
            return ''
        return ' from ' + ', '.join(filtered_columns)

    def _column(self, column, table):
        attributes = self._attributes(column['attributes'], table, True)
        return attributes

    def _attributes(self, attributes, source_alias, is_root):
        # assert len(attributes) == 1, f'Can only insert a single attribute per column, found: {attributes}'
        from_clauses = [self._attribute(attribute, source_alias, is_root) for attribute in attributes]
        return ', '.join([c for c in from_clauses if c])

    def _attribute(self, attribute, source_alias, is_root_table):
        # nothing to join if not a foreign key
        if not 'foreign-key' in attribute:
            return None

        # if foreign key is an extension on the root table, then we must not join if it's an insert query
        if is_root_table and attribute['foreign-key'].get('extension') and self._is_insert_query:
            return None

        from_clause = ''

        # if not root, then assume we need a left join. This may require a modifier at some point.
        if not is_root_table:
            from_clause += ' left join '

        # get foreign key
        foreign_key = attribute['foreign-key']

        # add target table
        from_clause += foreign_key['table']

        # add alias
        target_name = foreign_key['table']
        target_alias = foreign_key.get('alias', target_name)
        if target_alias != target_name:
            from_clause += f' as {target_alias}'

        # if not root then add on clause
        if not is_root_table:
            from_clause += f' on {source_alias}.{attribute["name"]} = {target_alias}.{foreign_key["name"]}'

            # if this is an Odoo style extension relation and it's not on the root table, then we need to filter by qualifier (module) and table (model)
            if foreign_key.get('extension'):
                qualifier = foreign_key['qualifier']
                # assume for now that source alias is the table name. This is fine as long as we're not joining the same table multiple times
                table_name = SelectCompiler().get_model_name(source_alias)
                from_clause += f' and {target_alias}.model = \'{table_name}\' and {target_alias}.module = \'{qualifier}\''


        # recurse
        return from_clause + self._attributes(foreign_key['attributes'], target_alias, False)



class ForeignWhereClauseCompiler:

    def __init__(self, is_insert_query):
        # if this is an insert query, we must not join with extension tables
        self._is_insert_query = is_insert_query
        self._aliases = {}

    def compile(self, mapping):
        # glue cells together
        table_name = mapping['table']
        clauses = [self._column(c, table_name) for c in mapping['columns']]

        return ' and '.join(chain(*clauses))

    def _column(self, column, table):
        return self._attributes(column['attributes'], table, True)

    def _attributes(self, attributes, source_alias, is_root):
        return chain(*[self._attribute(attribute, source_alias, is_root) for attribute in attributes])

    def _attribute(self, attribute, alias, is_root_table):
        if not 'foreign-key' in attribute:

            # only add where clauses for joined tables, but do register for alias
            if is_root_table:
                return []
            parameter_name = attribute.get('parameter', f'{attribute["name"]}')
            return [f'{alias}.{attribute["name"]} = :{parameter_name}']

        foreign_key = attribute['foreign-key']
        target_name = foreign_key['table']
        target_alias = foreign_key.get('alias', target_name)

        # don't add extension conditions if this is an insert query
        if is_root_table and foreign_key.get('extension') and self._is_insert_query:
            return []

        # recurse
        where_clause = self._attributes(foreign_key['attributes'], target_alias, False)

        # add extension conditions if this is the root of an update query
        if is_root_table and foreign_key.get('extension') and not self._is_insert_query:
            # assume for now that the alias is the table name. This is fine as long as we're not joining the same table multiple times
            model = alias.replace('_', '.')
            extension_clause = [f"{target_alias}.module = '{foreign_key['qualifier']}'",
                                f"{target_alias}.model = '{model}'"]
            where_clause = chain(where_clause, extension_clause)

        return where_clause


class ReturningClauseCompiler:
    # returns 'returning id' if this mapping contains an extension relation on the root table

    def compile(self, mapping):
        # get list of returned columns
        clauses = chain(*[self._column(c) for c in mapping['columns']])
        # filter out all empty clauses
        non_empty_clauses = [c for c in clauses if c]

        if not any(non_empty_clauses):
            return ''

        # assert there's no more than one
        assert len(non_empty_clauses) == 1, f'Can only return a single extension id, found: {non_empty_clauses}'

        return ' returning ' + mapping['table'] + '.' + non_empty_clauses[0]

    def _column(self, column):
        return self._attributes(column.get('attributes', []))

    def _attributes(self, attributes):
        return [self._attribute(attribute) for attribute in attributes]

    def _attribute(self, attribute):
        if not 'foreign-key' in attribute:
            # not a foreign key relationship, so not an extension
            return None
        if not attribute['foreign-key'].get('extension', False):
            # not an extension foreign key
            return None
        # extension, so return the id to return from the query. Default to 'id' bec/ that's what Odoo uses
        return attribute['foreign-key'].get('id', 'id')




'''
This module provides file sources to read data from.
'''
import os
import pickle
import re

import gspread
import pandas as pd
from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import InstalledAppFlow

from stimula.cli.file_source import StmlEvaluator

# If modifying these SCOPES, delete the file token.pickle.
SCOPES = ['https://www.googleapis.com/auth/spreadsheets.readonly']


def google_authenticate(credentials_file: str):
    # get folder name
    assert os.path.exists(credentials_file), f'Credentials file {credentials_file} not found in folder {os.getcwd()}. Please provide a valid Google credentials file. Browse to https://console.cloud.google.com/apis/credentials to create one.'
    flow = InstalledAppFlow.from_client_secrets_file(credentials_file, SCOPES)
    creds = flow.run_local_server(port=8080)

    # Save the credentials for the next run
    with open('token.pickle', 'wb') as token:
        pickle.dump(creds, token)


class GoogleSource:
    def __init__(self, sheet_id: str):
        self.sheet_id = sheet_id

    def google_get_credentials(self):
        # The file token.pickle stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first time.
        assert os.path.exists('token.pickle'), 'No token.pickle file found. Authenticate first.'
        with open('token.pickle', 'rb') as token:
            creds = pickle.load(token)

        assert creds, 'File token.pickle found but no valid credentials available. Authenticate first.'
        # If there are no valid credentials available, let the user log in.
        if not creds.valid and creds.expired and creds.refresh_token:
            creds.refresh(Request())

        # Save the credentials for the next run
        with open('token.pickle', 'wb') as token:
            pickle.dump(creds, token)

        return creds

    def read_files(self, file_names, table_names, context):
        creds = self.google_get_credentials()
        client = gspread.authorize(creds)
        # open the spreadsheet
        sheet = client.open_by_key(self.sheet_id)
        # create map from sheet name to sheet
        all_sheets = {ws.title: ws for ws in sheet.worksheets()}
        # get file_names that do not exist in all_sheets
        missing_files = set(file_names) - set(all_sheets.keys())
        assert not missing_files, f"Sheet(s) {', '.join(missing_files)} not found in spreadsheet."
        # get all worksheets
        sheets = [all_sheets[file_name] for file_name in file_names]
        # read contents of each sheet into a dataframe
        sheets_as_df = [pd.DataFrame(ws.get_all_values()) for ws in sheets]

        # derive table names from file names if not provided
        table_names = table_names or [self._table_name_from_file_name(file_name) for file_name in file_names]
        # derive context from file names if not provided
        context = context or file_names

        # instantiate STML evaluator with lambda to read file from disk
        stml_evaluator = StmlEvaluator(lambda _, source_file_name: self._read_file(all_sheets, source_file_name))

        # replace STML files with their source
        file_contents, table_names, context = stml_evaluator.replace_stmls_with_sources(file_names, sheets_as_df, table_names, context)

        return [file_contents, table_names, context]

    def _table_name_from_file_name(self, file_name):
        # remove all characters after the first character that is not a letter or digit, or underscore
        return re.sub(r'[^a-zA-Z0-9_].*$', '', file_name)

    def _read_file(self, all_sheets, source_file_name):
        values = all_sheets[source_file_name].get_all_values()
        df = pd.DataFrame(values[1:], columns=values[0])
        return df



"""
This is an abstract base class for authentication. It provides methods to authenticate and validate tokens.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
import base64
import os
import time

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from abc import ABC, abstractmethod


class Auth(ABC):
    # set the secret key during instantiation
    def __init__(self, secret_key_function, lifetime_function=lambda database: 24 * 3600):
        # secret_key function must be set.
        # We can't evaluate the secret key in the constructor because we don't have a database connection during instantiation.
        assert secret_key_function, 'Secret key must be set'
        self._secret_key_function = secret_key_function
        self._lifetime_function = lifetime_function

    def get_database_and_username(self, token):
        # decode token, without verifying the signature
        payload = jwt.decode(token, options={"verify_signature": False})
        # return database and username for easy re-authentication
        return payload.get('database'), payload.get('username')

    def authenticate(self, database, username, password):
        # validate submitted credentials and obtain user ID to store in token
        uid = self._validate_submitted_credentials(database, username, password)

        # evaluate secret key function for this database
        secret_key = self._secret_key_function(database)

        # encrypt password
        encrypted_password, salt = self.encrypt(secret_key, password)

        # issued at
        iat = int(time.time())

        # create token payload
        payload = {"database": database, "username": username, "uid": uid, "password": encrypted_password, "salt": salt, "iat": iat}
        token = jwt.encode(payload, secret_key, algorithm='HS256')
        return token

    def validate_token(self, token):
        # get database from token
        database, username = self.get_database_and_username(token)

        # evaluate secret key function for this database
        secret_key = self._secret_key_function(database)

        # validate and decode token
        payload = jwt.decode(token, secret_key, algorithms=['HS256'])

        # get connection parameters from payload
        database, uid, encrypted_password, salt, iat = payload['database'], payload['uid'], payload['password'], payload['salt'], payload['iat']

        # check if token is expired
        lifetime = self._lifetime_function(database)
        if iat + lifetime <= int(time.time()):
            raise jwt.ExpiredSignatureError('Token has expired, please log in again.')

        # decrypt password
        password = self.decrypt(secret_key, encrypted_password, salt)

        # validate token credentials and return resulting objects
        cnx, cr = self._validate_token_credentials(database, uid, password)

        # return validation result
        return cnx, cr, username

    @abstractmethod
    def _validate_submitted_credentials(self, database, username, password):
        pass

    @abstractmethod
    def _validate_token_credentials(self, database, username, password):
        pass

    def encrypt(self, secret, plaintext):
        salt = os.urandom(16)  # Generate a random salt
        dek = self._derive_key_from_secret(secret, salt)

        encrypted_secret = self._encrypt_with_derived_key(plaintext, dek)
        return base64.urlsafe_b64encode(encrypted_secret).decode(), base64.urlsafe_b64encode(salt).decode()

    def _encrypt_with_derived_key(self, plaintext, dek):
        cipher = Cipher(algorithms.AES(dek), modes.ECB(), backend=default_backend())
        encryptor = cipher.encryptor()

        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        padded_data = padder.update(plaintext.encode()) + padder.finalize()

        ciphertext = encryptor.update(padded_data) + encryptor.finalize()
        return ciphertext

    def decrypt(self, secret, ciphertext, salt):
        dek = self._derive_key_from_secret(secret, base64.urlsafe_b64decode(salt))
        encrypted_data = base64.urlsafe_b64decode(ciphertext)

        decrypted_data = self._decrypt_with_derived_key(encrypted_data, dek)

        return decrypted_data

    def _decrypt_with_derived_key(self, ciphertext, dek):
        cipher = Cipher(algorithms.AES(dek), modes.ECB(), backend=default_backend())
        decryptor = cipher.decryptor()
        decrypted_data = decryptor.update(ciphertext) + decryptor.finalize()

        # Use PKCS7 unpadding
        unpadder = padding.PKCS7(algorithms.AES.block_size).unpadder()
        plaintext = unpadder.update(decrypted_data) + unpadder.finalize()

        return plaintext.decode()

    def _derive_key_from_secret(self, passphrase, salt):
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,  # Length of the derived key
            salt=salt,
            iterations=100000,  # Adjust as needed
            backend=default_backend()
        )
        # return base64.urlsafe_b64encode(kdf.derive(passphrase.encode())).decode()
        return kdf.derive(passphrase.encode())



from .sly import Lexer


class HeaderLexer(Lexer):
    # Set of token names.   This is always required
    tokens = {ID, VALUE, QUOTED_VALUE, COLON, EQUALS, LPAREN, RPAREN, LBRACK, RBRACK}

    # String containing ignored characters between tokens
    ignore = ' \t'

    # Regular expression rules for tokens
    ID = r'[a-zA-Z_][a-zA-Z0-9_-]*'
    VALUE = r'[a-zA-Z0-9_$]+'
    QUOTED_VALUE = r'"[^\"]+"'
    COLON = r'\:'
    EQUALS = r'\='
    LPAREN = r'\('
    RPAREN = r'\)'
    LBRACK = r'\['
    RBRACK = r'\]'



"""
This script provides the abstract SqlCreator and concrete classes for Insert, Update, and Delete SQL creators.

Turning a diff into SQL statements works as follows:

1. Iterate rows in diff, because each query depends on the values in a row. Then repeat the following steps for each row.
2. Create a dictionary with unique column headers as keys and values as values. We'll need these for all query types.
3. Create a dictionary with other column headers as keys and values as values. Include non-null values for inserts and modified values for updates. Leave it empty for deletes.
4. Filter tree based on keys in the two dictionaries.
5. Compile the filtered tree to get parameter names for the query.
6. For each column in the tree, look up values by header using the dictionaries created earlier
7. Split values if the column has more than one parameter names. A CSV cell can contain multiple values separated by a colon
7. Clean up values. Strip whitespace from strings. Convert values to match the DB schema. For example, convert '1' to 1 if the column is an int
8. Create the dictionary with parameter names and values.
9. Compile the filtered tree to get the query for a row. Use the parameter dictionary where needed. For example, in SQL you write 'x = 1' but 'x is null'
10. Return query and parameter dictionary

Author: Romke Jonker
Email: romke@rnadesign.net
"""

import numpy
import pandas as pd
from psycopg2._json import Json

from .query_executor import SimpleQueryExecutor, DependentQueryExecutor, OperationType, FailedQueryExecutor
from ..compiler.delete_compiler import DeleteCompiler
from ..compiler.header_compiler import HeaderCompiler
from ..compiler.insert_compiler import InsertCompiler, ReturningClauseCompiler
from ..compiler.parameters_compiler import ParametersCompiler
from ..compiler.update_compiler import UpdateCompiler
from ..header.values_parser import ValuesLexer, ValuesParser


class SqlCreator:

    def __init__(self):
        self._values_lexer = ValuesLexer()
        self._values_parser = ValuesParser()
        self.operation_type = None

    def create_sql(self, mapping, diffs, context=None):

        # iterate rows in diff
        for i in range(len(diffs)):
            row = diffs.iloc[i]

            # return line number, depends on operation type
            line_number = self._get_line_number(row)

            try:
                # create query for row
                operation_type, query, value_dict = self._create_sql_row(mapping, row)

                # yield query and split columns
                yield SimpleQueryExecutor(line_number, operation_type, mapping['table'], query, value_dict, context)
            except Exception as e:
                # yield query with line number and error message
                yield FailedQueryExecutor(line_number, self.operation_type, mapping['table'], context, str(e))


    def _create_sql_row(self, mapping, row):
        # Create a dictionary with unique column headers as keys and values as values. We'll need these for all query types.
        unique_value_dict = self._create_unique_value_dict(mapping, row)

        # raise error if there are no unique values
        assert unique_value_dict, 'Header must have at least one unique column'

        # Create a dictionary with other column headers as keys and values as values.
        non_unique_value_dict = self._create_non_unique_value_dict(mapping, row)

        # create combined dictionary for convenience
        header_value_dict = {**unique_value_dict, **non_unique_value_dict}

        # Filter tree based on keys in the combined dictionary
        filtered_mapping = self._filter_mapping(mapping, header_value_dict)

        # Compile the filtered tree to get parameter names for the query.
        parameter_names = ParametersCompiler().compile(filtered_mapping)

        # Create the dictionary with parameter names as keys and values as values. An element may contain multiple parameters and values
        parameter_value_dict = self._map_parameter_names_with_values(filtered_mapping, parameter_names, header_value_dict)

        # Split values if the column has more than one parameter names. A CSV cell can contain multiple values separated by a colon
        split_parameter_value_dict = self._split_columns(parameter_names, parameter_value_dict)

        # Clean up values. Strip whitespace from strings. Convert values to match the DB schema. For example, convert '1' to 1 if the column is an int
        value_dict_clean = self._clean_values_in_dict(split_parameter_value_dict)

        # Compile the filtered tree to get the query for a row. Use the parameter dictionary where needed. For example, in SQL you write 'x = 1' but 'x is null'
        operation_type, query = self._create_query(filtered_mapping, value_dict_clean)

        return operation_type, query, value_dict_clean

    def _create_unique_value_dict(self, mapping, row):
        # get unique column headers
        unique_headers = HeaderCompiler().compile_list_unique(mapping)

        # create dictionary with unique column headers as keys and values as values
        unique_value_dict = {header: row[header] for header in unique_headers}

        return unique_value_dict

    def _create_non_unique_value_dict(self, mapping, row):
        # get non-unique headers
        non_unique_headers = HeaderCompiler().compile_list_non_unique(mapping)

        # create dictionary with non-unique column headers as keys and values as values
        non_unique_value_dict = {header: row[header] for header in non_unique_headers}

        return non_unique_value_dict

    def _create_non_unique_root_extension_value_dict(self, mapping, row):
        # when we delete a record with an extension on the root table (not on a foreign key table),
        # then we need to delete the extension record as well. If the extension is unique, then
        # the delete statement will already have the extension's name value, but if it's not unique,
        # then this method ensures the value is included.

        # get non-unique headers
        non_unique_headers = HeaderCompiler().compile_list_non_unique_root_extension(mapping)

        # create dictionary with non-unique column headers as keys and values as values
        non_unique_value_dict = {header: row[header] for header in non_unique_headers}

        return non_unique_value_dict

    def _filter_mapping(self, mapping, value_dict):
        # filter columns by those that have a value in value_dict

        # create a list of column headers
        headers = HeaderCompiler().compile_list(mapping, include_skip=True)

        # create a list of all columns
        all_columns = mapping['columns']

        assert len(headers) == len(all_columns), f'Number of headers must equal number of columns, found: {len(headers)} and {len(all_columns)}'

        # create a list of items in mapping where the corresponding header is in value_dict
        filter_columns = [item for item, header in zip(all_columns, headers) if header in value_dict]

        # create a copy of mapping with columns replaced by filtered columns
        filtered_mapping = mapping.copy()
        filtered_mapping['columns'] = filter_columns

        return filtered_mapping

    def _map_parameter_names_with_values(self, filtered_mapping, parameter_names, value_dict):
        # create a list of column headers
        headers = HeaderCompiler().compile_list(filtered_mapping)

        # create a dictionary with parameter names as keys and values as values
        return {parameter_name: value_dict[header] for parameter_name, header in zip(parameter_names, headers)}

    def _split_columns(self, parameter_names, value_dict):
        # verify that parameter_names and value_dict have the same number of items
        assert len(parameter_names) == len(value_dict), f'Number of parameter names must equal number of values, found: {parameter_names} and {value_dict}'

        # create empty dictionary
        value_dict_split = {}

        # iterate parameter names
        for parameter_name in parameter_names:
            # split the value and add it to the dictionary
            value_dict_split.update(self._split_value(parameter_name, value_dict[parameter_name]))

        return value_dict_split

    def _split_value(self, names, values):
        # if there is only one name, there is only one value
        if len(names) == 1:
            return {names[0]: values}

        # else split values at colon, but make sure not to split at json colons
        split_values = self._values_parser.parse(self._values_lexer.tokenize(values))

        # verify that number of names equals number of values
        assert len(names) == len(split_values), f'Number of names must equal number of values, found: {names} and {split_values}'

        # create dictionary with parameter names as keys and values as values
        return {name: value for name, value in zip(names, split_values)}

    def _clean_values_in_dict(self, value_dict):
        return {key: self._clean_value_in_dict(value) for key, value in value_dict.items()}

    def _clean_value_in_dict(self, value):
        # strip whitespace from strings only
        if isinstance(value, str):
            value = value.strip()

        # convert numpy.int64 to int
        if isinstance(value, numpy.int64):
            value = int(value)

        # convert numpy.float64 to float
        if isinstance(value, numpy.float64):
            value = float(value)

        # convert numpy bool to python bool
        if isinstance(value, numpy.bool_):
            # psycopg wants string representations for bools
            value = 'true' if value else 'false'

        # convert dict to json string
        if isinstance(value, dict):
            # use psycopg's Json to correctly convert unicode characters
            value = Json(value)

        return value

    def _create_query(self, tree, value_dict):
        # raise exception because this method must be overridden
        raise NotImplementedError


class InsertSqlCreator(SqlCreator):
    def __init__(self):
        super().__init__()
        self.operation_type = OperationType.INSERT

    def create_sql(self, mapping, diffs, context=None):
        # override to create dependent insert queries for extensions
        for query_executor in super().create_sql(mapping, diffs, context):

            # is there an extension on the root table?
            if not ReturningClauseCompiler().compile(mapping):
                yield query_executor

            else:

                # create a separate query for the extension
                dependent_query = self._create_extension_insert_query(mapping, query_executor.params)
                dependent_query_executor = DependentQueryExecutor(query_executor.line_number, query_executor.operation_type, mapping['table'], context, (query_executor.query, query_executor.params),
                                                                  dependent_query)
                yield dependent_query_executor

    def _create_non_unique_value_dict(self, mapping, row):
        # call super to get all other columns and values
        non_unique_value_dict = super()._create_non_unique_value_dict(mapping, row)

        # remove null and nan values
        non_unique_value_dict = {key: value for key, value in non_unique_value_dict.items() if self._has_value(value)}

        return non_unique_value_dict

    def _has_value(self, value):
        return value is not None and not pd.isnull(value) and not pd.isna(value) and not value == ''

    def _create_query(self, mapping, value_dict):
        # create insert query
        return self.operation_type, InsertCompiler().compile(mapping)

    def _create_extension_insert_query(self, mapping, param):
        # get table, name parameter name and values
        table, name_parameter_name, values = ExtensionValueHelper().get_extension_parameter_values(mapping, param)

        # create insert query
        sql = f'insert into {table} (name, module, model, res_id) values (:{name_parameter_name}, :module, :model, :res_id)'
        return sql, values

    def _get_line_number(self, row):
        # for inserts, just return __line__ column
        return row['__line__']


class UpdateSqlCreator(SqlCreator):

    def __init__(self):
        super().__init__()
        self.operation_type = OperationType.UPDATE

    def _create_unique_value_dict(self, mapping, row):
        # split row in self and other
        self_row, other_row = self._split_diff_self_other(row)

        # get unique headers
        unique_headers = HeaderCompiler().compile_list_unique(mapping)

        # create dictionary with unique column headers as keys and values as values
        self_unique_value_dict = {header: self_row[header] for header in unique_headers}
        other_unique_value_dict = {header: other_row[header] for header in unique_headers}

        # validate that values for self and other are the same, we can't change unique values
        assert self_unique_value_dict == other_unique_value_dict, f'Values for unique columns must be the same for self and other, found: {self_unique_value_dict} and {other_unique_value_dict}'

        return self_unique_value_dict

    def _create_non_unique_value_dict(self, mapping, row):
        # split row in self and other
        self_row, other_row = self._split_diff_self_other(row)

        # get unique columns
        unique_headers = HeaderCompiler().compile_list_unique(mapping)

        # get self columns by removing unique headers from self row keys
        non_unique_headers = [header for header in self_row.keys() if header not in unique_headers]

        # create new result dictionary
        non_unique_value_dict = {}

        # iterate non-unique self headers
        for header in non_unique_headers:

            # get self and other values
            self_value = self_row[header]
            other_value = other_row[header]

            # if self and other values are different, add them to the result dictionary
            if self._is_value_modified(self_value, other_value):
                # these values may be <NA>. We need to convert them to None before passing them to the DB, but
                # better to keep them here as <NA> bec/ pandas doesn't like None values. The constructor converts None to <NA>
                # in our test cases.
                non_unique_value_dict[header] = self_value

        # verify that there is at least one modified value
        assert len(non_unique_value_dict) > 0, f'At least one non-unique value must be modified, found: {len(non_unique_value_dict)}'

        return non_unique_value_dict

    def _split_diff_self_other(self, row):
        # extract self and other columns into a new Series
        self_row = row[[column for column in row.index if (column[0] != '__line__' and column[1] in ['', 'self'])]]
        other_row = row[[column for column in row.index if (column[0] != '__line__' and column[1] in ['', 'other'])]]

        # replace header tuples with original column names
        self_row.index = [column[0] for column in self_row.index]
        other_row.index = [column[0] for column in other_row.index]

        # return self and other rows
        return self_row, other_row

    def _is_value_modified(self, this, that):
        # if both values are null, then it's not modified
        if self._is_empty(this) and self._is_empty(that):
            return False

        # if one value is null and the other value is not null, it's modified
        if self._is_empty(this) != self._is_empty(that):
            return True

        # else compare non-null values
        return this != that

    def _is_empty(self, value):
        return value is None or pd.isnull(value) or pd.isna(value) or value == ''

    def _create_query(self, tree, value_dict):
        # create query
        return self.operation_type, UpdateCompiler().compile(tree)

    def _get_line_number(self, row):
        # for updates, the column contains Series, not sure how to change that
        return row['__line__'].iloc[0]


class DeleteSqlCreator(SqlCreator):
    def __init__(self):
        super().__init__()
        self.operation_type = OperationType.DELETE


    def create_sql(self, mapping, diffs, context=None):
        # override to create dependent delete queries for extensions
        for query_executor in super().create_sql(mapping, diffs, context):

            # is there an extension on the root table?
            if not ReturningClauseCompiler().compile(mapping):
                yield query_executor

            else:
                # create a separate query for the extension
                dependent_query = self._create_extension_delete_query(mapping, query_executor.params)
                dependent_query_executor = DependentQueryExecutor(query_executor.line_number, query_executor.operation_type, mapping['table'], context, (query_executor.query, query_executor.params),
                                                                  dependent_query)
                yield dependent_query_executor

    def _create_non_unique_value_dict(self, mapping, row):
        # For deletes we don't need non-unique columns.
        # Except, we do need extension name value, because we need to delete the extension even if it's not unique
        return super()._create_non_unique_root_extension_value_dict(mapping, row)

    def _create_query(self, mapping, value_dict):
        # create query
        return self.operation_type, DeleteCompiler().compile(mapping)

    def _get_line_number(self, row):
        # deletes don't have a matching input row
        return None

    def _create_extension_delete_query(self, mapping, param):
        # get table, name parameter name and values
        table, name_parameter_name, values = ExtensionValueHelper().get_extension_parameter_values(mapping, param)

        # create delete query
        sql = f'delete from {table} where name = :{name_parameter_name} and module = :module and model = :model and res_id = :res_id'
        return sql, values


class ExtensionValueHelper:
    def get_extension_parameter_values(self, mapping, param):
        # get extension foreign keys
        extension_foreign_keys = [a['foreign-key'] for c in mapping['columns'] for a in c.get('attributes', []) if a.get('foreign-key', {}).get('extension')]

        # only one extension is supported for now
        assert len(extension_foreign_keys) == 1, f'Only one extension is supported, found: {len(extension_foreign_keys)}'
        foreign_key = extension_foreign_keys[0]

        # foreign key must have a single attribute
        assert len(foreign_key['attributes']) == 1, f'Foreign key must have a single attribute, found: {len(foreign_key["attributes"])}'
        attribute = foreign_key['attributes'][0]
        name_parameter_name = attribute['parameter']

        table = foreign_key['table']
        name = param[name_parameter_name]
        qualifier = foreign_key['qualifier']
        # TODO: replace with proper way to get the model name
        model = mapping['table'].replace('_', '.')

        # create values map
        values = {name_parameter_name: name, 'module': qualifier, 'model': model, 'res_id': None}

        # return table, name parameter name and values
        return table, name_parameter_name, values



"""
This script defines an Invoker class intended for remote invocations of the Stimula API.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
import jwt
import requests


class Invoker:
    def __init__(self, remote):
        self._remote = remote

    def set_context(self, token):
        # save token for consecutive requests
        self._token = token

    def auth(self, database, username, password):
        # create connection url
        path = 'auth'

        # post credentials as form data
        data = {"database": database, "username": username, "password": password}

        # return the token from json response
        return self.post(path, data=data, send_token=False).json()['token']

    def get_database_and_username(self, token):
        # decode token, without verifying the signature
        payload = jwt.decode(token, options={"verify_signature": False})
        # return database and username for easy re-authentication
        return payload['database'], payload['username']

    def list(self, filter):
        # set path
        path = 'tables'

        # send filter as query parameter
        params = {"q": filter}

        # return the token from json response
        return self.get(path, params).json()

    def mapping(self, filter):
        # create connection url
        path = f"tables/{filter}/header"

        # send filter as query parameter
        params = {"style": "csv"}

        # return the token from json response
        return self.get(path, params).text

    def count(self, table, header, query):
        # create connection url
        path = f"tables/{table}/count"

        # send filter as query parameter
        params = {"h": header, "q": query}

        # return the token from json response
        return self.get(path, params).json()['count']

    def get_table(self, table, header, query):
        path = f"tables/{table}"

        # send filter as query parameter
        params = {"h": header, "q": query}

        # return the token from json response
        return self.get(path, params).text

    def post_table(self, tables, header, query, files, skiprows, insert, update, delete, execute, commit, format, post_script, context):
        assert files is not None and len(files) > 0, 'Provide one or more files to post'

        assert len(tables) == len(files), "Provide exactly one file per table, not %s" % len(files)

        if len(files) == 1:
            # post single file from disk or stdin
            path = f"tables/{tables[0]}"

            # create query parameters
            params = {"h": header, "q": query, 'skiprows': skiprows, 'insert': insert, 'update': update, 'delete': delete, 'execute': execute, 'commit': commit,
                      'style': format}

            # return the result from json response
            return self.post(path, params=params, data=files[0]).json()

        else:
            # post multiple files
            path = f"tables"

            # send filter as query parameter
            params = {'t': ','.join(tables), 'h': header, 'insert': insert, 'update': update, 'delete': delete, 'execute': execute, 'commit': commit}

            # zip table names and files to create file dictionary for post request. Make sure the keys are unique
            file_map = {f'file{suffix}': (file_name, file, 'text/csv') for suffix, file_name, file in zip(range(len(files)), context, files)}

            # return the token from json response
            return self.post_multi(path, params, files=file_map).json()

    def get(self, path, params):
        # create connection url
        url = f"{self._remote}/stimula/1.0/{path}"

        # set bearer token
        headers = {"Authorization": f"Bearer {self._token}"}

        # get table list from the url
        response = requests.get(url, headers=headers, params=params)

        # if the response is not 200, raise an exception
        if response.status_code != 200:
            raise Exception(f"Request failed: {response.json()['msg']}\nRemote trace: {response.json()['trace']}")

        # return the response
        return response

    def post(self, path, params=None, data=None, send_token=True):
        # create connection url
        url = f"{self._remote}/stimula/1.0/{path}"

        headers = {}

        # set bearer token
        if send_token:
            headers["Authorization"] = f"Bearer {self._token}"

        # post data to the url
        response = requests.post(url, headers=headers, params=params, data=data)

        # if the response is not 200, raise an exception
        if response.status_code != 200:
            self._raise_error(response)

        return response

    def post_multi(self, path, params=None, files=None):
        # create connection url
        url = f"{self._remote}/stimula/1.0/{path}"

        # set bearer token
        headers = {"Authorization": f"Bearer {self._token}"}

        # Send the POST request with the files dictionary
        response = requests.post(url, headers=headers, params=params, files=files)

        # if the response is not 200, raise an exception
        if response.status_code != 200:
            self._raise_error(response)

        return response

    def _raise_error(self, response):
        try:
            response_json = response.json()
        except:
            raise Exception(f"Request failed: {response.text}")
        raise Exception(f"Request failed: {response_json.get('msg')}\nRemote trace: {response_json.get('trace')}")
